(******************************************************************
 *                                                                *
 * Cryptographic protocol verifier                                *
 *                                                                *
 * Bruno Blanchet, David Cadé, Charlie Jacomme, and Benjamin Lipp *
 *                                                                *
 * Copyright (C) ENS, CNRS, INRIA, MPI-SP, 2005-2023              *
 *                                                                *
 ******************************************************************)

(*

    Copyright ENS, CNRS, INRIA, MPI-SP
    contributors: Bruno Blanchet, Bruno.Blanchet@inria.fr
                  David Cadé
                  Charlie Jacomme, Charlie.Jacomme@inria.fr
                  Benjamin Lipp, benjamin.lipp@mpi-sp.org

This software is a computer program whose purpose is to verify 
cryptographic protocols in the computational model.

This software is governed by the CeCILL-B license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-B
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-B license and that you accept its terms.

*)
module NatMap

module Map = FStar.Map

type t value = nat * Map.t nat value

let next_id (#value: Type) (m: t value) = match m with | nid, m -> nid
let map (#value: Type) (m: t value) = match m with | nid, m -> m

let sel t n = Map.sel (map t) n

let upd t n v = next_id t, Map.upd (map t) n v

let empty v =
  let m = Map.const v in
  0, Map.restrict Set.empty m

let domain t = Map.domain (map t)

let contains t id = Map.contains (map t) id

let reserve t = match t with | nid, m -> nid, ((nid + 1), m)

let add t v = match t with | nid, m -> nid, ((nid + 1), Map.upd m nid v)

let remove t id =
  match t with
  | nid, m ->
    let new_dom = Set.complement #nat (Set.singleton #nat id) in
    nid, Map.restrict #nat new_dom m

let get m id = match m with | nid, m -> if Map.contains m id then Some (Map.sel m id) else None

